% Copyright (c) 2016, NVIDIA CORPORATION. All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%  * Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%  * Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in the
%    documentation and/or other materials provided with the distribution.
%  * Neither the name of NVIDIA CORPORATION nor the names of its
%    contributors may be used to endorse or promote products derived
%    from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
% PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%
%Description: s-step Conjugate Gradient (CG) to solve linear system Ax=f
%Input Parameters:
%s - number of steps to be bundled together
%A - s.p.d. coefficient matrix of the linear system
%f - right-hand-side of the linear system
%tol - convergence tolerance based on relative residual
%mit - maximum number of iterations allowed 
%x0  - initial guess
%Output Parameters:
%x - approximation to the solution of the linear system
%r - residual based on x
%i - iteration at which the method stopped
%h - residual history

function  [x,r,i,h] = s_step_cg(s,A,f,tol,mit,x0)
%error checking
if s < 1
   fprintf(1,'%s\n','Error s<1') 
end

%initial setup
x = x0;
r = f-A*x;
r0= r;

%start main loop
for i=1:mit
    %s-step matrix-vector multiply
    %T_nxs+1  = [r,(A*inv(M))^1*r,(A*inv(M))^2*r,...,(A*inv(M))^s*r]
    T=s_matvecs(A,r,s+1);
    R = T(:,1:s);    
    AR= T(:,2:s+1);
    
    %find directions p
    if i == 1
        P=R;
    else                
        %block-dot product            
        C = -AR'*P;
        %find scalars beta
        B = W\C;       
        P = R+P*B;
    end    
    
    %block-dot product
    W = AR'*P;
    g = P'*r; 
    %find scalars alpha
    a = W\g;

    %compute new approximation
    x = x + P*a;
    r = f-A*x;      
    
    %check residual norm    
    h(i) = norm(r);
    if (norm(r) < tol*norm(r0))
        break;
    end    
end

