% Copyright (c) 2016, NVIDIA CORPORATION. All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%  * Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%  * Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in the
%    documentation and/or other materials provided with the distribution.
%  * Neither the name of NVIDIA CORPORATION nor the names of its
%    contributors may be used to endorse or promote products derived
%    from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
% PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%
% Description: preconditioned s-step BiCGStab to solve linear system Ax=f
% Input Parameters:
% s - number of steps to be bundled together
% A - coefficient matrix of the linear system
% f - right-hand-side of the linear system
% tol - convergence tolerance based on relative residual
% mit - maximum number of iterations allowed 
% M   - preconditioner M=LU, such that M^-1 ~ A^-1
% x0  - initial guess
% Output Parameters:
% x - approximation to the solution of the linear system
% r - residual based on x
% i - iteration at which the method stopped
% hh- residual history

function [x,r,i,hh] = s_step_pbicgstab(s,A,f,tol,mit,L,U,x0)
%error checking
if s < 1
   fprintf(1,'%s\n','Error s<1') 
end
%breakdown parameters
bflag= 0;

%initial setup
[m,n]=size(A);
r0=f-A*x0;
%Choose either "rh=r0;" or "rh=(L*U)'*r0;", the latter needed if comparing 
%with standard BiCGStab on (A*inv(M))*(M*x) = f
%rh=r0;
rh=(L*U)'*r0;  
r =r0;
p =r0;
x =x0;

%pre-allocation for speed
a=zeros(s,1);
b=zeros(s,1);
o=zeros(s,1);
S=zeros(m,s);
Q=zeros(m,s);
V=zeros(m,s-1);
Z=zeros(m,s-1);

%main loop
for i=1:mit
    e = 2*s+1;
    %s-step matrix-vector multiply     
    %T_mx4s+2=       [r,p,(A*inv(M))*r,(A*inv(M))*p,...,(A*inv(M))^{2*s}*r,(A*inv(M))^{2*s}*p]
    %Y_mx4s  =inv(M)*[r,p,(A*inv(M))*r,(A*inv(M))*p,...,(A*inv(M))^{2*s-1}*r,(A*inv(M))^{2*s-1}*p] 
    [T,Y] = s_lumatvecs(A,L,U,[r,p],e);
    %extract R, P, V and Z from T
    for j=1:e
        k = 1+2*(j-1);
        R(:,j) = T(:,k); 
        P(:,j) = T(:,k+1);        
    end
    for j=1:e-1
        k = 1+2*(j-1);
        V(:,j) = Y(:,k); 
        Z(:,j) = Y(:,k+1);        
    end 
    %block dot-products
    W = [R,P]'*[R,P];
    g = rh'*V;
    h = rh'*Z;
 
    %inner loop
    for j=1:s
        %set index k
        k = 1+2*(j-1);
                
        %find alpha = rh'*r / rh'*A*p        
        a(j)=g(k)/h(k+1);
        
        %find S (R = [ri,S])
        %set inv(M)*S=inv(M)*[s, As, ..., A^{2*s-1}s]
        V(:,k+1:e-1) = V(:,k:e-2) - a(j)*Z(:,k+1:e-1);
        R(:,k+1:e)   = R(:,k:e-1) - a(j)*P(:,k+1:e);
        %Update g and W
        g(:,k+1:e-1) = g(:,k:e-2) - a(j)*h(:,k+1:e-1);
        W(:,k+1:e)   = W(:,k:e-1) - a(j)*W(:,e+k+1:2*e);
        W(k+1:e,:)   = W(k:e-1,:) - a(j)*W(e+k+1:2*e,:);
        
        %Check ||s||_{2} for breakdown
        if (sqrt(W(k+1,k+1)) < tol*norm(r0))
              bflag=1;
              break;
        end
        
        %find omega = s'*A*s / (A*s)'*(A*s)       
        o(j)=W(k+1,k+2)/W(k+2,k+2);  
        
        %find Q (P = [pi,Q])
        %set inv(M)*Q=inv(M)*[q, Aq, ..., A^{2*s-1}q]
        Z(:,k+1:e-1)  = Z(:,k:e-2)     - o(j)*Z(:,k+1:e-1);
        P(:,k+1:e)    = P(:,k:e-1)     - o(j)*P(:,k+1:e);
        %Update h and W
        h(:,k+1:e-1)  = h(:,k:e-2)     - o(j)*h(:,k+1:e-1);
        W(:,e+k+1:2*e)= W(:,e+k:2*e-1) - o(j)*W(:,e+k+1:2*e);
        W(e+k+1:2*e,:)= W(e+k:2*e-1,:) - o(j)*W(e+k+1:2*e,:);
        
        %WARNING: no need to compute r, q, and p in the last iteration,
        %residual r will be recomputed manually every s iterations.
        if (j == s)
            break;
        end            
        
        %find R (R = [ri,si,R_new])
        %set inv(M)*R_new=inv(M)*[r_new, Ar_new, ..., A^{2*(s-1)}r_new]        
        V(:,k+2:e-1) = V(:,k+1:e-2) - o(j)*V(:,k+2:e-1);
        R(:,k+2:e)   = R(:,k+1:e-1) - o(j)*R(:,k+2:e);
        %Update g and W
        g(:,k+2:e-1) = g(:,k+1:e-2) - o(j)*g(:,k+2:e-1);
        W(:,k+2:e)   = W(:,k+1:e-1)   - o(j)*W(:,k+2:e);
        W(k+2:e,:)   = W(k+1:e-1,:)   - o(j)*W(k+2:e,:);           
        
        %find beta = (rh'*r_new / rh'*r) * (alpha / omega)
        b(j)=(g(k+2)/g(k))*(a(j)/o(j));
                                             
        %find P (P = [pi,qi,P_new])
        %set inv(M)*P_new=inv(M)*=[p_new, Ap_new, ..., A^{2*(s-1)}p_new] 
        Z(:,k+2:e-1)  = V(:,k+2:e-1) + b(j)*Z(:,k+1:e-2);
        P(:,k+2:e)    = R(:,k+2:e)   + b(j)*P(:,k+1:e-1);
        %Update h and W
        h(:,k+2:e-1)  = g(:,k+2:e-1) + b(j)*h(:,k+1:e-2);
        W(:,e+k+2:2*e)= W(:,k+2:e)   + b(j)*W(:,e+k+1:2*e-1);
        W(e+k+2:2*e,:)= W(k+2:e,:)   + b(j)*W(e+k+1:2*e-1,:);
    end
     
    %WARNING: computing x requires one old p direction, and all new s directions. 
    %extract all constructed residual and directions
    for jj=1:j        
        k = 1+2*(jj-1);
        PP(:,jj) = Z(:,k); 
        SS(:,jj) = V(:,k+1); 
    end
    %if there was a breakdown, do not use last direction s
    if (bflag == 1)
        PP=PP(:,1:j);
        a = a(1:j);
        if (j == 1)
            SS=zeros(m,1);
            o =0;
        else
            SS=SS(:,1:j-1);
            o = o(1:j-1);
        end
    end
    
    %compute new approximation
    x = x + PP*a + SS*o;     
    r = f-A*x;  
    
    
    %check convergence
    hh(i) = norm(r);
    if ((norm(r) < tol*norm(r0)) | (bflag > 0))
        break
    end
        
    %compute residual preconditioned from the left (for the next iteration)
    mr= U\(L\r);
    %compute new beta and new direction p for the next iteration    
    k = 1+2*(s-1); %set index k when j=s 
    b(j) = ((rh'*mr)/g(k))*(a(j)/o(j));
    p = r + b(j)*P(:,k+1); %q = P(:,k+1) = PP(:,j) - o(j)*A*PP(:,j)
end

