% Copyright (c) 2016, NVIDIA CORPORATION. All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%  * Redistributions of source code must retain the above copyright
%    notice, this list of conditions and the following disclaimer.
%  * Redistributions in binary form must reproduce the above copyright
%    notice, this list of conditions and the following disclaimer in the
%    documentation and/or other materials provided with the distribution.
%  * Neither the name of NVIDIA CORPORATION nor the names of its
%    contributors may be used to endorse or promote products derived
%    from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
% EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
% IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
% PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
% CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
% EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
% PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
% PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
% OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
% (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
% OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
%
% Description: standard BiConjugate Gradient Stabilized (BiCGStab) to solve linear system Ax=f
% Input Parameters:
% A - coefficient matrix of the linear system
% f - right-hand-side of the linear system
% tol - convergence tolerance based on relative residual
% mit - maximum number of iterations allowed 
% x0  - initial guess
% Output Parameters:
% x - approximation to the solution of the linear system
% r - residual based on x
% i - iteration at which the method stopped
% h - residual history

function [x,r,i,h] = standard_bicgstab(A,f,tol,mit,x0)
%initial setup
[m,n]=size(A);
r0=f-A*x0;
x=x0;
r=r0;
p=r;
z=r; %z=r_hat

%main loop
for i=1:mit
    q=A*p;
    alpha = (z'*r)/(z'*q);      
    s = r - alpha*q;
    if (norm(s) < tol*norm(r0))
        x = x+alpha*p;
        r = f-A*x;
        h(i) = norm(r);
        i=i-0.5;
        break
    end
    t = A*s;
    omega = (s'*t)/(t'*t);
    x = x+alpha*p+omega*s;
    rp= r;    
    r = s-omega*t;
    beta = ((z'*r)/(z'*rp))*(alpha/omega);
    p = r + beta*(p-omega*q);
    h(i) = norm(r);
    if (norm(r) < tol*norm(r0))
        break
    end
end

